###### Libraries #############

library(lubridate)
library(ggplot2)
require(cowplot)
library(plyr)
library(readr)
library(tidyverse) 
library(scales)
library(zoo)
theme_set(theme_gray())


########## FUNCTIONS ##############################

# For computing month, month day, hour - cycle
getYearlyCycle <- function(df){
  
  DF <- df %>% 
    group_by(month, mday, hour) %>% 
    summarise(E = mean(amount),m = n()) %>% 
    ungroup()
  
  DF$m <- DF$m/sum(DF$m)
  
  DF$E = DF$E - weighted.mean(x = DF$E, w = DF$m , na.rm = TRUE)
  
  v <- unlist(apply(df, MARGIN = 1, function(x) DF[ (DF$month == as.integer(x['month'])  & DF$mday == as.integer(x['mday'])  & DF$hour ==  as.integer( x['hour'] ) ), "E" ] )) 
  
  return(v)
}

# For computing week of the year, day of the week  - cycle
getNWeekCycle <- function(df){
  
  DF <- df %>% 
    group_by(week, wday) %>% 
    summarise(E = mean(amount),m = n()) %>% 
    ungroup()
  
  DF$m <- DF$m/sum(DF$m)
  
  DF$E = DF$E - weighted.mean(x = DF$E, w = DF$m , na.rm = TRUE)
  
  v <- unlist(apply(df, MARGIN = 1, function(x) DF[ (DF$week == as.integer(x['week']) & DF$wday == as.integer(x['wday'])), "E" ] ))
  
  return(v)
}

# For computing day of the week, hour  - cycle
getWeeklyCycle <- function(df){
  
  DF <- df %>% 
    group_by(wday, hour) %>% 
    summarise(E = mean(amount),m = n()) %>% 
    ungroup()
  
  DF$m <- DF$m/sum(DF$m)
  
  DF$E = DF$E - weighted.mean(x = DF$E, w = DF$m , na.rm = TRUE)
  
  v <- unlist(apply(df, MARGIN = 1, function(x) DF[ (DF$wday == as.integer(x['wday']) & DF$hour ==  as.integer( x['hour'] ) ), "E" ] ))
  
  return(v)
}

# For preparing data to a useful table
Days <- function(x){
  days <- data.frame(hour = paste(x['dd/mm/aa'], '0', sep = ' '), cars =  x[2])
  for(i in 3:25){
    days <- rbind(days ,data.frame( hour = paste(x['dd/mm/aa'], as.character(i-2), sep = ' '),cars =  x[i]) )
  }
  return(days)
}

# For forecasting with the studiend seasonal/trend effects
Pred <- function(x) {
  S_C1 <- C1[C1$wday == as.integer(x['wday']) & C1$hour == as.integer(x['hour']), ]$S_Week
  S_C2 <- C2[ C2$week == as.integer(x['week']) & C2$wday == as.integer(x['wday']) & C2$hour == as.integer(x['hour']) , ]$S_NWeek
  S_C3 <- C3[ C3$month == as.integer(x['month']) & C3$mday == as.integer(x['mday']) & C3$hour == as.integer(x['hour']) , ]$S_Year
  Tr <- predict(trend, newdata =  data.frame(time = as.POSIXct(x['time'])) )
  if(length(S_C2)==0){
    S_C2 <- mean(C2[ C2$week == as.integer(x['week']) & C2$wday == (as.integer(x['wday']) + 5) , ]$S_NWeek)
  }
  result <- exp(Tr[[1]] + S_C1 + S_C2 + S_C3)
  return(result)
}

#For computing the minimum number of gates needed acording to Criteria 1
nGates.EL <- function(AR,SR, L){
  
  EL <- function( rho, available, L ){
    if(rho < 1){
      c <- available
      
      if(c == 1){
        
        result <- rho/(1-rho)
      }else{
        
        G = ((c*rho)**c)/((1-rho)*factorial(c))
        
        for( n in 0:(c-1) ){
          
          G = G + ((c*rho)**n)/(factorial(n))
        }
        
        result <- (rho*(c*rho)**c)/(factorial(c)*G*(1-rho)**2)
      }
    }else{
      result <- L*available + 1
    }
    return(result)
  }
  
  c = 0
  continue <-  TRUE
  
  while(continue){
    
    c <- c + 1
    rho <- AR/(SR*c)
    
    if( EL( rho , available = c, L) <= (L*c) ){
      continue <- FALSE
    }
    
  }
  return(data.frame( SERVERS_EL = c, expected_length = EL( rho , available = c, L )/c  ))
}

#For computing the minimum number of gates needed acording to Criteria 2
nGates.PL <- function(AR , SR , L, p, limit){
  PL <- function(rho, c, L, p){
    if(rho < 1){  
      if(c == 1){
        
        result <- rho**L
        
      }else{
        
        G <-  ((c*rho)**c)/((1-rho)*factorial(c))
        for( n in 0:(c-1) ){
          G <-  G + ((c*rho)**n)/(factorial(n))
        }
        
        result <- 1.
        
        if(L <= c){
          
          for( n in 0:(L-1) ){
            result <-  result - ((c*rho)**n)/(G*factorial(n))
          }
          
        }else{
          
          for( n in 0:(c-1) ){
            result <-  result - ((c*rho)**n)/(G*factorial(n))
          }
          
          for( n in c:(L-1) ){
            result <-  result - (c**c*rho**n)/(G*factorial(c))
          }
          
        }
      }
    }else{
      result <- 2
    }
    return(result)
  }
  
  c = 0
  continue <-  TRUE
  
  while(continue){
    
    c <- c + 1
    rho <- AR/(SR*c)
    
    if((PL( rho,  c, L*c, p) < p)){
      continue = FALSE
    }
    if(c > limit){
      return(data.frame (Number_Servers = 50, Probability = PL(rho, c, L*c, p) ))
    }
  }
  return(data.frame (Number_Servers = c, Probability = PL(rho, c, L*c, p) ))
}




## LOAD DATA  ##############################

df <- read_delim("~/Desktop/Initial_Data.csv",";", escape_double = FALSE, trim_ws = TRUE)

###### PREPARE DATA ##############################

# Select the ones in the N direction
df <- df[complete.cases(df),]

# Drop column of Direcctions
df <- df[ ,  !(names(df) %in% c('Dir.','T.dia') ) ]

#set hours of days as a unic column
dummy <- apply(df, 1, function(x) Days(x))
df <- dummy[[1]]

for(i in 2:length(dummy)){
  df <- rbind(df, dummy[[i]])
}

# remove dummy list
rm(dummy)

# reset row index
rownames(df) <- NULL

# change columns names
colnames(df) <- c('time', 'amount')

# arrange formats
df$time <- as.POSIXct(as.character(df$time),format = '%m/%d/%y %H' )
df$amount <- as.numeric(as.character(df$amount))

# Plot of initial data
ggplot(df, aes(x=df$time, y=df$amount)) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("hourly arrivals")

# Splitting data set
train<- df[ !((month(df$time) %in% c(10,11,12)) & year(df$time) == 2017)  ,]
test <- df[ ((month(df$time) %in% c(10,11,12)) & year(df$time) == 2017)  ,]

#Plot training data set
ggplot(train, aes(x=train$time, y=train$amount)) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)")

##### LOG TRANSFORMATION ######################

# Asssessing positivity for the log-transformation and do it.
train$amount[train$amount == 0] <- 1.
train$amount <- log( train$amount )

# Initial training data
ggplot(train, aes(x=train$time, y=train$amount)) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)")

# Detect, look for a patron and work on outliers
data.frame(day = day(train[train$amount < 2,'time']), month = month(train[train$amount < 2,'time']), hour = hour(train[train$amount < 2,'time']) , weekday =  weekdays(train[train$amount < 2,'time']))
df_dummy <- train[month(train$time) == 3 & hour(train$time) == 2 & weekdays(train$time) == "Sunday" & train$amount != 0 ,]
df_dummy$year <- year(df_dummy$time)
train[month(train$time) == 3 & hour(train$time) == 2 & weekdays(train$time) == "Sunday" & train$amount == 0 , 'amount'] <- unique(ave(df_dummy$amount, df_dummy$year ,FUN = mean))

# Remove df_dumy
rm(df_dummy)

# Plot data without outliers
ggplot(train, aes(x=train$time, y=train$amount)) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("date") + ylab("Ln(hourly arrivals)")


###  TREND ##############################

# Isolating the trend - plot
v1 <- rollmean(x = train$amount,k = 2000)
p1 <- qplot(seq_along(v1), v1, ylab = "", xlab = "", size = I(0.05), main = 'Sliding window = 2000')
v2 <- rollmean(x = train$amount,k = 5000) 
p2 <- qplot(seq_along(v2), v2 , ylab = "", xlab = "", size = I(0.05), main = 'Sliding window = 5000' )
v3 <- rollmean(x = train$amount,k = 10000) 
p3 <- qplot(seq_along(v3), v3, ylab = "", xlab = "", size = I(0.05), main = 'Sliding window = 10000')
v4 <- rollmean(x = train$amount,k = 20000) 
p4 <- qplot(seq_along(v4), v4, ylab = "", xlab = "", size = I(0.05), main = 'Sliding window = 20000') 
v5 <- rollmean(x = train$amount,k = 30000) 
p5 <- qplot(seq_along(v5), v5, ylab = "", xlab = "", size = I(0.05), main = 'Sliding window = 30000') 
v6 <- rollmean(x = train$amount,k = 40000) 
p6 <- qplot(seq_along(v6), v6, ylab = "", xlab = "", size = I(0.05), main = 'Sliding window = 40000') 
plot_grid(p1,p2,p3,p4,p5,p6, ncol = 3)

# freeing memory
rm(v1,v2,v3,v4,v5,v6,p1,p2,p3,p4,p5,p6)

# Compute linear approximation of the trend
trend <- lm(amount ~ time, train )

# Extract the trend from the TS
train$amount <- train$amount -  predict(trend, data.frame(time = train$time))

# Plot TS
ggplot(train, aes(x=train$time, y=train$amount)) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln( hourly arrivals )")

### SEASONAL ###################


######## Construct columns 
DescTimes <- as.POSIXlt(train$time, format="%m/%d/%y %H")

train$hour <- DescTimes$hour
train$wday <- DescTimes$wday
train$month <- DescTimes$mon 
train$mday <- DescTimes$mday -1
train$week <- as.numeric(strftime(DescTimes,format="%W"))
train$year <- year(DescTimes) 

######### STUDY  

#### C_1

# Computation hourly x weekday cycle
train$S_Week <- getWeeklyCycle(train) 

# Plot of the cycle
Amonth <- train[train$month == 6 & train$year == 2015,]
p1 <- ggplot(Amonth, aes(x=Amonth$time, y=(Amonth$S_Week))) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%b %d, %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)") + ggtitle("Single month")
p2 <- ggplot(train, aes(x=train$time, y=(train$S_Week))) + geom_line(size = 0.05) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)") + ggtitle("Whole training range")
plot_grid(p1,p2)

# Removing Amonth
rm(Amonth)

# Plot of TS after seasonal effect extraction
ggplot(train, aes(x=train$time, y=(train$amount-train$S_Week))) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)")

# Extraction of seasonal effect
train$amount <- train$amount - train$S_Week

#### C_2

# Computation of year-week/ day/ hour effect
train$S_NWeek <- getNWeekCycle(train)

# Plot of the cycle
ggplot(train, aes(x=train$time, y=(train$S_NWeek))) + geom_line(size = 0.05) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)")

# Plot of TS after seasonal effect extraction
ggplot(train, aes(x=train$time, y=(train$amount - train$S_NWeek))) + geom_line(size = 0.05) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)")

# Extraction of seasonal effect
train$amount <- train$amount - train$S_NWeek

#### C_3

# Computation of the month/month-day and hour effect
train$S_Year <- getYearlyCycle(train) # h md mon

# Plot of the cycle
ggplot(train, aes(x=train$time, y=(train$S_Year))) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)")

# Plot of TS after seasonal effect extraction
ggplot(train, aes(x=train$time, y=(train$amount - train$S_Year))) + geom_line(size = 0.1) + scale_x_datetime(labels = date_format("%b - %Y")) + geom_smooth(method = lm)  + xlab("time") + ylab("Ln(hourly arrivals)")

# Extraction of seasonal effect
train$amount <- train$amount - train$S_Year

# Extract off-peaks
train %>% 
  filter(amount < (-1.15)) %>% 
  select(time, wday, week)

### TESTING #################

#Construct tables of cyclical effects
C1 <- train%>% 
  group_by(wday, hour) %>% 
  summarise(S_Week = unique(S_Week)) %>% 
  ungroup()

C2 <- train%>% 
  group_by(week, wday, hour) %>% 
  summarise(S_NWeek = mean(S_NWeek)) %>% 
  ungroup()


C3 <- train%>% 
  group_by(month, mday, hour) %>% 
  summarise(S_Year = unique(S_Year)) %>% 
  ungroup()

# Construct columns for filtering
DescTimes <- as.POSIXlt(test$time, format="%m/%d/%y %H")

test$hour <- DescTimes$hour
test$wday <- DescTimes$wday
test$month <- DescTimes$mon 
test$mday <- DescTimes$mday -1
test$week <- as.numeric(strftime(DescTimes,format="%W"))

# Forcasting for each hour
test$prediction <- apply(test, MARGIN = 1, function(x) Pred(x))

# Plot forcasting vs real
df1 <- rbind(data.frame(time = test$time,value = test$prediction, kind = rep("Prediction", nrow(test))),data.frame(time = test$time,value = test$amount, kind = rep("Real", nrow(test))))
df1$kind <- as.factor(df1$kind)

ggplot(df1, aes(x=df1$time, y=df1$value,group = df1$kind, colour = df1$kind)) + geom_line(size = 0.2) + scale_x_datetime(labels = date_format("%b - %Y"))  + xlab("time") + ylab("Ln(hourly arrivals)") +scale_colour_discrete(name  ="Kind of data") + scale_color_manual(values = c("red","black"))

# Freeing data
rm(df1,C1,C2,C3, DescTimes )

##### Computing TS Error
ASE <- mean((test$prediction-test$amount)**2)
sqrt(ASE)
sqrt(ASE)/(mean(test$amount))

## QUEUING THEORY ################

#### Prepearing measures

# Introducing Service rate per minute
SR <-  5

# Values for Criteria 1
TH_E <- 2

# Values for Criteria 2
TH_P_L <- 3
TH_P_p <- 0.15
limit_P <- 50

## Computations for predicted arrivals 

# Set Arrival rate per minute
AR <-  test$prediction/60 

# Arranging a table
predi <- data.frame(time = test$time, Arrival.rate = AR, Service_rate = SR)

# Setting as a column for the table with dates for Criteria 1
dummy <- apply(predi, MARGIN = 1, function(x) nGates.EL(AR = as.numeric(x["Arrival.rate"]),SR = as.numeric(x["Service_rate"]),L =  TH_E) )
results_E <- dummy[[1]]
for(i in 2:length(dummy)){
  results_E <- rbind(results_E, dummy[[i]])
}
predi <- cbind(predi, results_E)

# Setting as a column for the table with dates for Criteria 1
dummy <- apply(predi, MARGIN = 1, function(x) nGates.PL(AR = as.numeric(x["Arrival.rate"]),SR = as.numeric(x["Service_rate"]), L =  TH_P_L, p = TH_P_p, limit = limit_P ) )
results_P <- dummy[[1]]
for(i in 2:length(dummy)){
  results_P <- rbind(results_P, dummy[[i]])
}
predi <- cbind(predi, results_P)

## Computations for real arrivals

# Set Arrival rate per minute
AR <-  test$amount/60 

# Arranging a table
real <- data.frame(time = test$time, Arrival.rate = AR, Service_rate = SR)

# Setting as a column for the table with dates for Criteria 1
dummy <- apply(real, MARGIN = 1, function(x) nGates.EL(AR = as.numeric(x["Arrival.rate"]),SR = as.numeric(x["Service_rate"]),L =  TH_E) )
results_E <- dummy[[1]]
for(i in 2:length(dummy)){
  results_E <- rbind(results_E, dummy[[i]])
}
real <- cbind(real, results_E)

# Setting as a column for the table with dates for Criteria 2
dummy <- apply(real, MARGIN = 1, function(x) nGates.PL(AR = as.numeric(x["Arrival.rate"]),SR = as.numeric(x["Service_rate"]), L =  TH_P_L, p = TH_P_p, limit = limit_P ) )
results_P <- dummy[[1]]
for(i in 2:length(dummy)){
  results_P <- rbind(results_P, dummy[[i]])
}
real <- cbind(real, results_P)

# removing dummy list
rm(dummy)

# Ploting for Criteria 1 predicted vs real
df1 <- rbind(data.frame(time = predi$time , gates = predi$SERVERS_EL, kind = rep("Prediction", nrow(predi))),data.frame(time = real$time, gates = real$SERVERS_EL, kind = rep("Real", nrow(real))))
df1$kind <- as.factor(df1$kind)

ggplot(df1, aes(x=df1$time, y= as.factor(df1$gates), group = df1$kind, colour = df1$kind)) + geom_line(size = 0.5) + scale_x_datetime(labels = date_format(" %b,%d %Y"))  + xlab("time") + ylab("Number of Gates") + scale_colour_discrete(name  ="Kind of data")

# Ploting for Criteria 2 predicted vs real
df1 <- rbind(data.frame(time = predi$time , gates = predi$Number_Servers, kind = rep("Prediction", nrow(predi))),data.frame(time = real$time, gates = real$Number_Servers, kind = rep("Real", nrow(real))))
df1$kind <- as.factor(df1$kind)

ggplot(df1, aes(x=df1$time, y=as.factor(df1$gates), group = df1$kind, colour = df1$kind)) + geom_line(size = 0.5) + scale_x_datetime(labels = date_format(" %b,%d %Y"))  + xlab("time") + ylab("Number of Gates") + scale_colour_discrete(name  ="Kind of data")

# Error measure for Criteria 1
ASE <- mean((predi$SERVERS_EL-real$SERVERS_EL)**2)
sqrt(ASE)
sqrt(ASE)/(mean(real$SERVERS_EL))

# Error measure for Criteria 2
ASE <- mean((predi$Number_Servers-real$Number_Servers)**2)
sqrt(ASE)
sqrt(ASE)/(mean(real$Number_Servers))

#### FORCASTING ###################
### Data preparation ##########################

# Once format and table preparations have been done.

## log tranformation
df$amount[df$amount == 0] <- 1.
df$amount <- log( df$amount )

## Outliers
df_dummy <- df[month(df$time) == 3 & hour(df$time) == 2 & weekdays(df$time) == "Sunday" & df$amount != 0 ,]
df_dummy$year <- year(df_dummy$time)
df[month(df$time) == 3 & hour(df$time) == 2 & weekdays(df$time) == "Sunday" & df$amount == 0 , 'amount'] <- unique(ave(df_dummy$amount, df_dummy$year ,FUN = mean))
rm(df_dummy)

## trend
trend <- lm(amount ~ time, df )

df$amount <- df$amount -  predict(trend, data.frame(time = df$time))

## cycle 

DescTimes <- as.POSIXlt(df$time, format="%m/%d/%y %H")
df$hour <- DescTimes$hour
df$wday <- DescTimes$wday
df$month <- DescTimes$mon 
df$mday <- DescTimes$mday -1
df$week <- as.numeric(strftime(DescTimes,format="%W"))

df$S_Week <- getWeeklyCycle(df)

df$amount <- df$amount - df$S_Week

df$S_NWeek <- getNWeekCycle(df)

df$amount <- df$amount - df$S_NWeek

df$S_Year <- getYearlyCycle(df)

df$amount <- df$amount - df$S_Year

## Arrivals forecasting #############################

# Computing tables with the cyclic effects
C1 <- df%>% 
  group_by(wday, hour) %>% 
  summarise(S_Week = unique(S_Week)) %>% 
  ungroup()

C2 <- df%>% 
  group_by(week, wday, hour) %>% 
  summarise(S_NWeek = mean(S_NWeek)) %>% 
  ungroup()


C3 <- df%>% 
  group_by(month, mday, hour) %>% 
  summarise(S_Year = unique(S_Year)) %>% 
  ungroup()

# Build hourly time serie from '01 01 2018 0' until '02 28 2018 23'
start <- as.POSIXct('01 01 2018 0', format="%m %d %Y %H")
end <- as.POSIXct('02 28 2018 23', format="%m %d %Y %H")
prediction <- data.frame(time = seq(from = start, to = end, by='hour'))

#Build columns for filtering
DescTimes <- as.POSIXlt(prediction$time, format="%m/%d/%y %H")
prediction$hour <- DescTimes$hour
prediction$wday <- DescTimes$wday
prediction$month <- DescTimes$mon 
prediction$mday <- DescTimes$mday -1
prediction$week <- as.numeric(strftime(DescTimes,format="%W"))

# Forecasting TS
prediction$forecast <- apply(prediction, MARGIN = 1, function(x) Pred(x))

# Plotting forecasted TS
ggplot(prediction, aes(x=prediction$time, y=prediction$forecast), col = 'red') + geom_line(size = 0.2, col = 'red') + scale_x_datetime(labels = date_format("%d of %b %Y"))  + xlab("time") + ylab("Ln(hourly arrivals)") 

## Queuing measures forcasting ##############################

# Expected Service rate per minute
SR <-  5

# Values for Criteria 1
TH_E <- 2

# Values for Criteria 2
TH_P_L <- 3
TH_P_p <- 0.15
limit_P <- 50

#### Computation for gate's predictions

# Compute arrival rate per minute
AR <-  prediction$forecast/60 

# Arrange table
predi <- data.frame(time = prediction$time, Arrival.rate = AR, Service_rate = SR)

# Predict minimum number of gates per hour according to Criteria 1
dummy <- apply(predi, MARGIN = 1, function(x) nGates.EL(AR = as.numeric(x["Arrival.rate"]),SR = as.numeric(x["Service_rate"]),L =  TH_E) )
results_E <- dummy[[1]]
for(i in 2:length(dummy)){
  results_E <- rbind(results_E, dummy[[i]])
}
predi <- cbind(predi, results_E)

# Plot forecasted number of gates
ggplot(predi, aes(x=predi$time, y=predi$SERVERS_EL)) + geom_line(size = 0.5, col = 'red') + scale_x_datetime(labels = date_format(" %b of %d %Y"))  + xlab("time") + ylab("Number of Gates") 

# Predict minimum number of gates per hour according to Criteria 1
dummy <- apply(predi, MARGIN = 1, function(x) nGates.PL(AR = as.numeric(x["Arrival.rate"]),SR = as.numeric(x["Service_rate"]), L =  TH_P_L, p = TH_P_p, limit = limit_P ) )
results_P <- dummy[[1]]
for(i in 2:length(dummy)){
  results_P <- rbind(results_P, dummy[[i]])
}
predi <- cbind(predi, results_P)

# Plot forecasted number of gates
ggplot(predi, aes(x=predi$time, y=predi$Number_Servers)) + geom_line(size = 0.5, col = 'red') + scale_x_datetime(labels = date_format(" %b of %d %Y"))  + xlab("time") + ylab("Number of Gates") 




